@@ -1,9 +1,10 @@
 package Pandoc::Elements;
 use strict;
+## no critic (ProhibitNoStrict, ProhibitSubroutinePrototypes)
 use warnings;
-use 5.010;
+use 5.010001;
 
-our $VERSION = '0.28';
+our $VERSION = '0.29';
 
 use Carp;
 use JSON qw(decode_json);
@@ -11,6 +12,7 @@ use Scalar::Util qw(blessed reftype);
 use Pandoc::Walker qw(walk);
 use Pandoc::Version;
 
+
 our $PANDOC_VERSION; # a string like '1.16'
 $PANDOC_VERSION ||= eval { Pandoc::Version->new($ENV{PANDOC_VERSION}) };
 
@@ -150,6 +152,7 @@ foreach my $name ( keys %ELEMENTS ) {
     $parent = join ' ', map { "Pandoc::Document::$_" } $parent,
       map { 'AttributesRole' } grep { $_ eq 'attr' } @accessors;
 
+    ## no critic (ProhibitStringyEval)
     eval "package $class; our \@ISA = qw($parent);";
 
     *{ __PACKAGE__ . "::$name" } = Scalar::Util::set_prototype(
@@ -171,6 +174,7 @@ foreach my $name ( keys %ELEMENTS ) {
         ? " [ map { bless \$_, 'Pandoc::Document::$1' } \@{$member} ];"
         : " $member;";
         for ( split '/', $accessors[$i] ) {
+            ## no critic
             *{ $class . "::$_" } = eval "sub { $code }";
         }
     }
@@ -178,7 +182,7 @@ foreach my $name ( keys %ELEMENTS ) {
 
 sub element {
     my $name = shift;
-    no strict 'refs';
+    no strict 'refs'; #
     croak "undefined element" unless defined $name;
     croak "unknown element $name" unless $ELEMENTS{$name};
     &$name(@_);
@@ -315,6 +319,7 @@ sub pandoc_json($) {
     package Pandoc::Document;
     use strict;
     use Carp 'croak';
+    use Pandoc;
     our $VERSION = '0.04';
     our @ISA = ('Pandoc::Document::Element');
     sub blocks;
@@ -349,6 +354,56 @@ sub pandoc_json($) {
         return $self->{'pandoc-api-version'};
     }
     *pandoc_version = \&Pandoc::Elements::pandoc_version;
+    sub outline {
+        my ($self, $depth) = @_;
+        _sections( [@{$self->blocks}], $depth );
+    }
+    sub _sections {
+        my ($list, $depth) = @_;
+        my (@blocks, @sections);
+
+        # everything up to the first Header
+        while (@$list) {
+            if ($list->[0]->name eq 'Header') {
+                last if !$depth or $depth >= $list->[0]->level;
+            }
+            push @blocks, shift @$list;
+        }
+
+        # divide into sections
+        while (@$list) {
+            my $header = shift @$list;
+            my $level  = $header->level;
+
+            my @content;
+            while (@$list) {
+                if ($list->[0]->name eq 'Header') {
+                    last if $list->[0]->level <= $level;
+                }
+                push @content, shift @$list;
+            }
+
+            my $s = ($depth and $depth < $level)
+                ? { blocks => \@content }
+                : _sections(\@content,$depth);
+            push @sections, { header => $header, %$s };
+        }
+
+        return { blocks => \@blocks, sections => \@sections };
+    }
+    sub to_pandoc {
+        my ($self, @args) = @_;
+        my $pandoc = pandoc;
+        my $in = $self->to_json;
+        $pandoc->run( [ -f => 'json', @args ], { in => \$in, out => \my $out } );
+        return $out;
+    }
+    foreach my $format (qw(markdown latex html rst plain)) {
+        no strict 'refs';
+        *{ __PACKAGE__ . "::to_$format" } = sub {
+            shift()->to_pandoc( @_, '-t' => $format );
+        }
+    }
 }
 
 {
@@ -418,7 +473,7 @@ sub pandoc_json($) {
     sub string {
 
         # TODO: fix issue #4 to avoid this duplication
-        if ( $_[0]->name =~ /^(Str|Code|Math|MetaString)$/ ) {
+        if ( $_[0]->name =~ /^(Str|Code|CodeBlock|Math|MetaString)$/ ) {
             return $_[0]->content;
         }
         elsif ( $_[0]->name =~ /^(LineBreak|SoftBreak|Space)$/ ) {
@@ -427,7 +482,7 @@ sub pandoc_json($) {
         join '', @{
             $_[0]->query(
                 {
-                    'Str|Code|Math|MetaString'  => sub { $_->content },
+                    'Str|Code|CodeBlock|Math|MetaString'  => sub { $_->content },
                     'LineBreak|Space|SoftBreak' => sub { ' ' },
                 }
             );
@@ -483,7 +538,7 @@ sub pandoc_json($) {
     sub classes {
         my $e = shift;
         croak 'Method classes() is not a setter' if @_;
-        warn "->classes is deprecated, use ->class instead\n";
+        warn "->classes is deprecated. Use [ split ' ', \$e->class ] instead\n";
         $e->attr->[1]
     }
 
@@ -557,6 +612,10 @@ sub pandoc_json($) {
     our $VERSION = $PANDOC::Document::VERSION;
     our @ISA     = ('Pandoc::Document::Element');
     sub is_block { 1 }
+    sub null {
+        %{$_[0]} = (t => 'Null', c => []);
+        bless $_[0], 'Pandoc::Document::Null';
+    }
 }
 
 {
@@ -586,8 +645,8 @@ sub pandoc_json($) {
         unshift @{"Pandoc::Document::${Element}::ISA"}, __PACKAGE__; # no critic
     }
 
-    sub url   { $_[0]->{c}->[-1][0] }
-    sub title { $_[0]->{c}->[-1][1] }
+    sub url   { my $e = shift; $e->{c}->[-1][0] = shift if @_; return $e->{c}->[-1][0] //= ""; }
+    sub title { my $e = shift; $e->{c}->[-1][1] = shift if @_; return $e->{c}->[-1][1] //= ""; }
 
     sub upgrade {
         # prepend attributes to old-style ast
@@ -683,7 +742,7 @@ sub Pandoc::Document::MetaInlines::metavalue {
 }
 
 sub Pandoc::Document::MetaBlocks::metavalue {
-    join "\n", map { $_->string } @{$_[0]->{c}}
+    [ map { $_->string } @{$_[0]->{c}} ]
 }
 
 sub Pandoc::Document::MetaList::metavalue {
@@ -741,6 +800,7 @@ Pandoc::Elements - create and process Pandoc documents
 [![Build Status](https://travis-ci.org/nichtich/Pandoc-Elements.svg)](https://travis-ci.org/nichtich/Pandoc-Elements)
 [![Coverage Status](https://coveralls.io/repos/nichtich/Pandoc-Elements/badge.svg)](https://coveralls.io/r/nichtich/Pandoc-Elements)
 [![Kwalitee Score](http://cpants.cpanauthors.org/dist/Pandoc-Elements.png)](http://cpants.cpanauthors.org/dist/Pandoc-Elements)
+[![Code Climate Issue Count](https://codeclimate.com/github/nichtich/Pandoc-Elements/badges/issue_count.svg)](https://codeclimate.com/github/nichtich/Pandoc-Elements)
 
 =end markdown
 
@@ -1090,6 +1150,45 @@ ELEMENTS> flattened to unblessed values:
     $doc->metavalue   # equivalent to
     { map { $_ => $doc->meta->{$_}->metavalue } keys %{$doc->meta} }
 
+=item B<to_pandoc( [ @arguments ])>
+
+Process the document with L<Pandoc> executable and return its output:
+
+    $doc->to_pandoc( -o => 'doc.html' );
+    my $markdown = $doc->to_pandoc( -t => 'markdown' );
+
+=item B<to_...( [ @arguments ] )>
+
+Process the document into C<markdown> (pandoc's extended Markdown), C<latex>
+(LaTeX), C<html> (HTML), C<rst> (reStructuredText), or C<plain> (plain text).
+The following are equivalent:
+
+    $doc->to_markdown( @args );
+    $doc->to_pandoc( @args, '-t' => 'markdown' );
+
+=item B<outline( [ $depth ] )>
+
+Returns an outline of the document structure based on L<Header|/Header>
+elements. The outline is a hierarchical hash reference with the following
+fields:
+
+=over
+
+=item header
+
+L<Header|/Header> element (not included at the document root)
+
+=item blocks
+
+List of L<block elements|/Block elements> before the next L<Header|/Header>
+element (of given depth or less if a maximum depth was given)
+
+=item sections
+
+List of subsections, each having the same outline structure.
+
+=back
+
 =back
 
 =head2 BLOCK ELEMENTS
